@@ -30,7 +30,6 @@
 #include <ksba.h>
 
 #include "keydb.h"
-#include "../common/exechelp.h"
 #include "../common/i18n.h"
 #include "../common/sysutils.h"
 #include "minip12.h"
@@ -318,7 +317,7 @@ gpgsm_export (ctrl_t ctrl, strlist_t names, estream_t stream)
 
 /* Export a certificate and its private key.  RAWMODE controls the
    actual output:
-       0 - Private key and certifciate in PKCS#12 format
+       0 - Private key and certificate in PKCS#12 format
        1 - Only unencrypted private key in PKCS#8 format
        2 - Only unencrypted private key in PKCS#1 format
     */
@@ -430,7 +429,7 @@ gpgsm_p12_export (ctrl_t ctrl, const char *name, estream_t stream, int rawmode)
 
   if (rawmode == 0)
     ctrl->pem_name = "PKCS12";
-  else if (gpgsm_get_key_algo_info (cert, NULL) == GCRY_PK_ECC)
+  else if (gpgsm_get_key_algo_info (cert, NULL, NULL) == GCRY_PK_ECC)
     ctrl->pem_name = "EC PRIVATE KEY";
   else if (rawmode == 1)
     ctrl->pem_name = "PRIVATE KEY";
@@ -485,7 +484,7 @@ gpgsm_p12_export (ctrl_t ctrl, const char *name, estream_t stream, int rawmode)
 }
 
 
-/* Print some info about the certifciate CERT to FP or STREAM */
+/* Print some info about the certificate CERT to FP or STREAM */
 static void
 print_short_info (ksba_cert_t cert, estream_t stream)
 {
