@@ -500,6 +500,8 @@ static struct debug_flags_s debug_flags [] =
 static struct compatibility_flags_s compatibility_flags [] =
   {
     { COMPAT_ALLOW_KA_TO_ENCR, "allow-ka-to-encr" },
+    { COMPAT_NO_CHAIN_CACHE, "no-chain-cache"     },
+    { COMPAT_NO_KEYINFO_CACHE, "no-keyinfo-cache" },
     { 0, NULL }
   };
 
@@ -536,6 +538,9 @@ static int default_include_certs = DEFAULT_INCLUDE_CERTS;
 /* Whether the chain mode shall be used for validation.  */
 static int default_validation_model;
 
+/* Counter used to convey data from deinit_ctrl to gpgsm_exit.  */
+static unsigned int parent_cache_stats;
+
 /* The default cipher algo.  */
 #define DEFAULT_CIPHER_ALGO "AES256"
 
@@ -1798,6 +1803,10 @@ main ( int argc, char **argv)
       gnupg_inhibit_set_foregound_window (1);
     }
 
+  /* Better make sure that we have a statusfp so that a failure status
+   * in gpgsm_exit can work even w/o any preeding status messages. */
+  gpgsm_init_statusfp (&ctrl);
+
   /* Add default keybox. */
   if (!nrings && default_keyring && !opt.use_keyboxd)
     {
@@ -2351,9 +2360,16 @@ gpgsm_exit (int rc)
   else if (opt.assert_signer_list && !assert_signer_true)
     rc = 1;
 
+  /* If we had an error but not printed an error message, do it now.
+   * Note that the function will never print a second failure status
+   * line. */
+  if (rc)
+    gpgsm_exit_failure_status ();
+
   gcry_control (GCRYCTL_UPDATE_RANDOM_SEED_FILE);
   if (opt.debug & DBG_MEMSTAT_VALUE)
     {
+      log_info ("cert_chain_cache: cached=%u\n", parent_cache_stats);
       gcry_control( GCRYCTL_DUMP_MEMORY_STATS );
       gcry_control( GCRYCTL_DUMP_RANDOM_STATS );
     }
@@ -2381,9 +2397,23 @@ gpgsm_init_default_ctrl (struct server_control_s *ctrl)
 void
 gpgsm_deinit_default_ctrl (ctrl_t ctrl)
 {
+  unsigned int n;
+
   gpgsm_keydb_deinit_session_data (ctrl);
+  gpgsm_flush_keyinfo_cache (ctrl);
   xfree (ctrl->revocation_reason);
   ctrl->revocation_reason = NULL;
+  n = 0;
+  while (ctrl->parent_cert_cache)
+    {
+      cert_cache_item_t next = ctrl->parent_cert_cache->next;
+      ksba_cert_release (ctrl->parent_cert_cache->result);
+      xfree (ctrl->parent_cert_cache);
+      ctrl->parent_cert_cache = next;
+      n++;
+    }
+  if (n > parent_cache_stats)
+    parent_cache_stats = n;
 }
 
 
